@@ -20,6 +20,7 @@ from pathlib import Path
 from typing import Dict, Iterable, List, NamedTuple, Optional, Set, Tuple, cast
 
 import pkg_resources
+from semantic_version import Version
 from setuptools.command.build import build as CommandBuild  # type: ignore[import]
 from setuptools.command.build_ext import build_ext as CommandBuildExt
 from setuptools.command.build_ext import get_abi3_suffix
@@ -28,7 +29,21 @@ from typing_extensions import Literal
 from ._utils import format_called_process_error
 from .command import RustCommand
 from .extension import Binding, RustBin, RustExtension, Strip
-from .rustc_info import get_rust_host, get_rust_target_list, get_rustc_cfgs
+from .rustc_info import (
+    get_rust_host,
+    get_rust_target_list,
+    get_rust_version,
+    get_rustc_cfgs,
+)
+
+
+def _check_cargo_supports_crate_type_option() -> bool:
+    version = get_rust_version()
+
+    if version is None:
+        return False
+
+    return version.major > 1 or (version.major == 1 and version.minor >= 64)  # type: ignore
 
 
 class build_rust(RustCommand):
@@ -93,8 +108,10 @@ class build_rust(RustCommand):
 
     def get_data_dir(self) -> str:
         components = (
-            pkg_resources.safe_name(self.distribution.get_name()).replace("-", "_"),  # type: ignore[attr-defined]
-            pkg_resources.safe_version(self.distribution.get_version()).replace("-", "_"),  # type: ignore[attr-defined]
+            pkg_resources.safe_name(self.distribution.get_name()).replace("-", "_"),
+            pkg_resources.safe_version(self.distribution.get_version()).replace(
+                "-", "_"
+            ),
         )
         if self.build_number:
             components += (self.build_number,)
@@ -130,7 +147,6 @@ class build_rust(RustCommand):
     def build_extension(
         self, ext: RustExtension, forced_target_triple: Optional[str] = None
     ) -> List["_BuiltModule"]:
-
         target_triple = self._detect_rust_target(forced_target_triple)
         rustc_cfgs = get_rustc_cfgs(target_triple)
 
@@ -143,6 +159,7 @@ class build_rust(RustCommand):
 
         quiet = self.qbuild or ext.quiet
         debug = self._is_debug_build(ext)
+        use_cargo_crate_type = _check_cargo_supports_crate_type_option()
 
         cargo_args = self._cargo_args(
             ext=ext, target_triple=target_triple, release=not debug, quiet=quiet
@@ -161,11 +178,18 @@ class build_rust(RustCommand):
             ]
 
         else:
-            rustc_args = [
-                "--crate-type",
-                "cdylib",
-                *ext.rustc_flags,
-            ]
+            # If toolchain >= 1.64.0, use '--crate-type' option of cargo.
+            # See https://github.com/PyO3/setuptools-rust/issues/320
+            if use_cargo_crate_type:
+                rustc_args = [
+                    *ext.rustc_flags,
+                ]
+            else:
+                rustc_args = [
+                    "--crate-type",
+                    "cdylib",
+                    *ext.rustc_flags,
+                ]
 
             # OSX requires special linker arguments
             if rustc_cfgs.get("target_os") == "macos":
@@ -190,6 +214,9 @@ class build_rust(RustCommand):
             ):
                 rustc_args.extend(["-C", f"link-args=-sSIDE_MODULE=2 -sWASM_BIGINT"])
 
+            if use_cargo_crate_type and "--crate-type" not in cargo_args:
+                cargo_args.extend(["--crate-type", "cdylib"])
+
             command = [
                 self.cargo,
                 "rustc",
@@ -249,7 +276,7 @@ class build_rust(RustCommand):
             artifacts = _find_cargo_artifacts(
                 cargo_messages.splitlines(),
                 package_id=package_id,
-                kind="bin",
+                kinds={"bin"},
             )
             for name, dest in ext.target.items():
                 if not name:
@@ -277,15 +304,15 @@ class build_rust(RustCommand):
             artifacts = _find_cargo_artifacts(
                 cargo_messages.splitlines(),
                 package_id=package_id,
-                kind="cdylib",
+                kinds={"cdylib", "dylib"},
             )
             if len(artifacts) == 0:
                 raise DistutilsExecError(
-                    "Rust build failed; unable to find any build artifacts"
+                    "Rust build failed; unable to find any cdylib or dylib build artifacts"
                 )
             elif len(artifacts) > 1:
                 raise DistutilsExecError(
-                    f"Rust build failed; expected only one build artifact but found {artifacts}"
+                    f"Rust build failed; expected only one cdylib or dylib build artifact but found {artifacts}"
                 )
 
             artifact_path = artifacts[0]
@@ -345,7 +372,18 @@ class build_rust(RustCommand):
                 os.makedirs(os.path.dirname(ext_path), exist_ok=True)
 
             log.info("Copying rust artifact from %s to %s", dylib_path, ext_path)
-            shutil.copyfile(dylib_path, ext_path)
+
+            # We want to atomically replace any existing library file. We can't
+            # just copy the new library directly on top of the old one as that
+            # causes the existing library to be modified (rather the replaced).
+            # This means that any process that currently uses the shared library
+            # will see it modified and likely segfault.
+            #
+            # We first copy the file to the same directory, as `os.rename`
+            # doesn't work across file system boundaries.
+            temp_ext_path = ext_path + "~"
+            shutil.copyfile(dylib_path, temp_ext_path)
+            os.replace(temp_ext_path, ext_path)
 
             if sys.platform != "win32" and not debug_build:
                 args = []
@@ -657,7 +695,7 @@ def _find_cargo_artifacts(
     cargo_messages: List[str],
     *,
     package_id: str,
-    kind: str,
+    kinds: Set[str],
 ) -> List[str]:
     """Identifies cargo artifacts built for the given `package_id` from the
     provided cargo_messages.
@@ -666,11 +704,11 @@ def _find_cargo_artifacts(
     ...    [
     ...        '{"some_irrelevant_message": []}',
     ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib"]},"filenames":["/some/path/baz.so"]}',
-    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["dylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
     ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
     ...    ],
     ...    package_id="some_id",
-    ...    kind="cdylib",
+    ...    kinds={"cdylib", "dylib"},
     ... )
     ['/some/path/baz.so', '/file/two/baz.dylib']
     >>> _find_cargo_artifacts(
@@ -681,14 +719,14 @@ def _find_cargo_artifacts(
     ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
     ...    ],
     ...    package_id="some_id",
-    ...    kind="rlib",
+    ...    kinds={"rlib"},
     ... )
     ['/file/two/baz.rlib']
     """
     artifacts = []
     for message in cargo_messages:
         # only bother parsing messages that look like a match
-        if "compiler-artifact" in message and package_id in message and kind in message:
+        if "compiler-artifact" in message and package_id in message:
             parsed = json.loads(message)
             # verify the message is correct
             if (
@@ -698,7 +736,7 @@ def _find_cargo_artifacts(
                 for artifact_kind, filename in zip(
                     parsed["target"]["kind"], parsed["filenames"]
                 ):
-                    if artifact_kind == kind:
+                    if artifact_kind in kinds:
                         artifacts.append(filename)
     return artifacts
 
